/***********************************************
BETA.PRC
Useful procedures for beta-binomial-logit likelihood analysis
************************************************/

/************************************************
LnGamma
Computes the natural log of the gamma function
Uses Stirling's approximation.
Inputs:
	x	matrix argument
Outputs:
	g	ln(gamma(x))
The elements of x should be positive. 
For non-positive values, LnGamma returns values of "0"
*************************************************/

proc LnGamma(x);
local d,g,xd,xt,x2,x3,x5,x7,x9;
    d=(x.<=0);		@ non-positives @
    xd=x.*(1-d)+d;		@ converts non-positives to unity @
    xt=xd+2; 		@ to allow use of Stirling's approximation @
    x2 = xt.*xt;
    x3 = xt.*x2;
    x5 = x3.*x2;
    x7 = x5.*x2;
    x9 = x7.*x2;
     /* stirling's approximation */
    g = 0.5 * ln(2*pi) + (xt - 0.5) .* ln(xt) - xt + 1./(12.*xt) - 1./(360.*x3) 
        + 1./(1260.*x5) - 1./(1680.*x7) + 1./(1188.*x9);
    g=g-ln(xd.*(xd+1));	@ removes the added "2" @
    g=g.*(1-d);		@ returns "0" for non-positives @
retp(g);
endp;


/************************************************
D_LnGamma, D2_LnGamma,
Computes the first and second derivatives of the natural log of the gamma function.
Uses Stirling's approximation.
The elements of x should be positive. 
For non-positive values, the procs return values of "0"

*************************************************/
proc D_LnGamma(x);
local d,xt,g,xd,x2,x4,x6,x8,x10;
    d=(x.<=0);		@ non-positives @
    xd=x.*(1-d)+d;		@ converts non-positives to unity @
    xt=xd+2; 		@ to allow use of Stirling's approximation @
    x2 = xt.*xt;
    x4 = x2.*x2;
    x6 = x4.*x2;
    x8 = x6.*x2;
    x10 = x8.*x2;
    g = ln(xt) - 1./(2.*xt) - 1./(12.*x2) + 1./(120.*x4) - 1./(252.*x6) 
        + 1./(240.*x8) - 1./(132.*x10) ;
    g = g - 1./xd - 1./(1+xd);	@ removes the added "2" @
    g = g.*(1-d);		@ returns "0" for non-positives @
retp(g);
endp;

/************************************************/
proc D2_LnGamma(x);
local d,g,xd,xt,x2,x3,x5,x7,x9,x11;
    d=(x.<=0);		@ non-positives @
    xd=x.*(1-d)+d;		@ converts non-positives to unity @
    xt=xd+2; 		@ to allow use of Stirling's approximation @
    x2 = xt.*xt;
    x3 = xt.*x2;
    x5 = x3.*x2;
    x7 = x5.*x2;
    x9 = x7.*x2;
    x11 = x9.*x2;
    g = 1./xt + 1./(2.*x2) + 1./(6.*x3) - 1./(30.*x5) + 1./(42.*x7) - 1./(30.*x9) + 1./(11.*x11);
    g = g + 1./(xd.^2) + 1./((1+xd).^2);  @ removes the added "2" @
    g = g.*(1-d);		@ returns "0" for non-positives @
retp(g);
endp;

/************************************************
Like_Beta
Log Likelihood of Beta-Binomial-Logit Model

Input:	theta	parameter vector
Output:	negative log-likelihood

Takes _y _x and _w from the global variable table
_y={ya,yb} are the dependent variables
_x 	regressor matrix
_w 	vector indicating inclusion of the observation. 
	(1 for inclusion, 0 for exclusion)
	
*************************************************/
declare _y,_x,_w;

proc Like_Beta(theta);
local y,y1,y2,x,mu,s,sl,a,b,pp,w;
  w=varget("_w");   x=varget("_x"); 
  y=varget("_y").*w;  y1=y[.,1]; y2=y[.,2];
  mu=logit(x*theta[1:cols(x)]);
  s=w.*theta[rows(theta)]; sl=.01; s=(s-sl).*(s.>sl)+sl;
  a=mu.*s; b=(1-mu).*s;
  pp=LnGamma(y1+a)+LnGamma(y2+b)-LnGamma(y1+y2+s)-LnGamma(a)-LnGamma(b)+LnGamma(s);
retp(-sumc(pp));
endp;

/************************************************
G_Like_Beta
Gradient of Log Likelihood of Beta-Binomial-Logit Model

Input:	theta	parameter vector
Output:	1xk 	derivative vector evaluated at theta

Takes _y _x and _w from the global variable table
_y={ya,yb} are the dependent variables
_x 	regressor matrix
_w 	vector indicating inclusion of the observation. 
	(1 for inclusion, 0 for exclusion)
*************************************************/

proc G_Like_Beta(theta);
local gr;
  gr=(sumc(V_G_Like_Beta(theta,varget("_y"),varget("_x"))))';
retp(gr);
endp;

/************************************************
V_G_Like_Beta
Vector Gradient of Log Likelihood of Beta-Binomial-Logit Model

Inputs:	theta	parameter vector
	y	dependent variables, nx2
	x	regressor matrix
Output:	nxk 	derivative matrix evaluated at theta

Note: This is useful for calculation of robust covariance matrices
Takes _w from the global variable table
_w 	vector indicating inclusion of the observation. 
	(1 for inclusion, 0 for exclusion)
*************************************************/

proc V_G_Like_Beta(theta,y,x);
local mu,s,sl,a,b,dga,dgb,dgs,b1,b2,gr,w,y1,y2;
  w=varget("_w");   
  y1=y[.,1].*w;  y2=y[.,2].*w;
  mu=logit(x*theta[1:cols(x)]);
  s=w.*theta[rows(theta)]; sl=.01; s=(s-sl).*(s.>sl)+sl;
  a=mu.*s; b=(1-mu).*s;
  dga=D_LnGamma(y1+a)-D_LnGamma(a);    
  dgb=D_LnGamma(y2+b)-D_LnGamma(b); 
  dgs=D_LnGamma(y1+y2+s)-D_LnGamma(s);
  b1=(dga-dgb).*mu.*(1-mu).*s;
  b2=dga.*mu+dgb.*(1-mu)-dgs;
  gr=(x.*b1)~b2;
retp(-gr);
endp;

/************************************************
H_Like_Beta
Hessian of Log Likelihood of Beta-Binomial-Logit Model

Input:	theta	parameter vector
Output:	kxk	second derivative matrix evaluated at theta

Takes _y _x and _w from the global variable table
_y={ya,yb} are the dependent variables
_x 	regressor matrix
_w 	vector indicating inclusion of the observation. 
	(1 for inclusion, 0 for exclusion)
*************************************************/

proc H_Like_Beta(theta);
local y,y1,y2,x,rx,mu,s,sl,a,b,mu2,dga,dgb,dgs,dha,dhb,dhs,b11,b21,b22,h11,h22,h21,h,w;
  w=varget("_w");   
  y=varget("_y").*w; y1=y[.,1]; y2=y[.,2];
  x=varget("_x").*w;     rx=cols(x);
  mu=logit(x*theta[1:rx]);
  s=w.*theta[rows(theta)]; sl=.01; s=(s-sl).*(s.>sl)+sl;
  a=mu.*s; b=(1-mu).*s;
  mu2=mu.*(1-mu).*s;
  dga=D_LnGamma(y1+a)-D_LnGamma(a);    
  dgb=D_LnGamma(y2+b)-D_LnGamma(b); 
  dgs=D_LnGamma(y1+y2+s)-D_LnGamma(s);
  dha=D2_LnGamma(y1+a)-D2_LnGamma(a);    
  dhb=D2_LnGamma(y2+b)-D2_LnGamma(b); 
  dhs=D2_LnGamma(y1+y2+s)-D2_LnGamma(s);
  b11=(dga-dgb).*mu2.*(1-2*mu)+(dha+dhb).*(mu2.^2);
  b21=(dha.*mu.*s-dhb.*(1-mu).*s+dga-dgb).*mu.*(1-mu);
  b22=dha.*(mu.^2)+dhb.*((1-mu).^2)-dhs;
  h11=x'(x.*b11);
  h21=x'b21;
  h22=sumc(b22);
  h=(h11~h21)|((h21')~h22);
retp(-h);
endp;

/************************************************/

fn expt(x)=exp(x.*(x.>(-700)).*(x.<700)-(x.<=(-700))*700+(x.>700)*700);
fn logit(u)=1./(1+expt(-u));
